import uuid

import networkx as nx
import numpy as np
from flask import Flask, render_template, request, session

from zndraw import globals, io

app = Flask(__name__)
app.secret_key = str(uuid.uuid4())


@app.route("/")
def index():
    """Render the main ZnDraw page."""
    session["key"] = str(uuid.uuid4())  # TODO use session key e.g. for atoms cache
    return render_template("index.html", config=globals.config.dict())


@app.route("/config", methods=["POST", "GET"])
def config():
    """Get the zndraw configuration."""
    if request.method == "POST":
        print(f"Updating config: {request.json}")
        for key, value in request.json.items():
            setattr(globals.config, key, value)
    return {
        **globals.config.dict(),
        "total_frames": len(globals.config._atoms_cache) - 1,
    }


@app.route("/graph", methods=["POST"])
def get_graph():
    step = request.json
    try:
        atoms = globals.config.get_atoms(step=int(step))
        graph = io.get_graph(atoms)
        return {
            "nodes": [{**graph.nodes[idx], "id": idx} for idx in graph.nodes],
            "edges": list(graph.edges),
        }
    except KeyError:
        return {}


@app.route("/positions", methods=["POST"])
def positions_step():
    params = request.json
    result = []
    try:
        for step in range(params["start"], params["stop"]):
            atoms = globals.config.get_atoms(step=int(step))
            result.append(atoms.get_positions().tolist())
        return result
    except KeyError:
        return result


@app.route("/select", methods=["POST"])
def select() -> list[int]:
    """Update the selected atoms."""
    step = request.json["step"]
    method = request.json["method"]
    try:
        selected_ids = [int(x) for x in request.json["selected_ids"]]
    except TypeError:
        selected_ids = []
    if method in ["particles", "none"]:
        return {"selected_ids": selected_ids, "updated": False}
    elif method == "species":
        atoms = globals.config.get_atoms(step)

        for id in tuple(selected_ids):
            selected_symbol = atoms[id].symbol
            selected_ids += [
                idx for idx, atom in enumerate(atoms) if atom.symbol == selected_symbol
            ]
        return {"selected_ids": list(set(selected_ids)), "updated": True}
    elif method == "connected":
        atoms = globals.config.get_atoms(step)
        graph = io.get_graph(atoms)

        total_ids = []

        for node_id in selected_ids:
            total_ids += list(nx.node_connected_component(graph, node_id))

        return {"selected_ids": list(set(total_ids)), "updated": True}

    else:
        raise ValueError(f"Unknown selection method: {method}")


@app.route("/add_update_function", methods=["POST"])
def add_update_function():
    """Add a function to the config."""
    try:
        signature = globals.config.add_update_function(request.json)
    except (ImportError, ValueError) as err:
        return {"error": str(err)}
    return signature


@app.route("/set_update_function_parameter", methods=["POST"])
def set_update_function_parameter():
    """Update the values of the update function."""
    globals.config.set_update_function_parameter(request.json)
    return {}


@app.route("/select_update_function/<name>")
def select_update_function(name):
    """Select a function from the config."""
    if name == "none":
        name = None
    globals.config.active_update_function = name
    return {}


@app.route("/update", methods=["POST"])
def update_scene():
    selected_ids = list(sorted(request.json["selected_ids"]))
    step = request.json["step"]
    points = np.array([[x["x"], x["y"], x["z"]] for x in request.json["points"]])
    globals.config.apply_update_function(selected_ids, step, points=points)
    return {}


@app.route("/load")
def load():
    """Function to call asynchronously to load atoms in the background."""
    globals.config.load_atoms()
    return {}



import contextlib
import pathlib
import socket
import sys
import webbrowser

import typer

from zndraw import __version__, app, globals

try:
    import webview as wv
except ImportError:
    wv = None

cli = typer.Typer()


def version_callback():
    typer.echo(f"ZnDraw {__version__}")
    raise typer.Exit()


@cli.command()
def main(
    file: str = typer.Argument(..., help="Trajectory File"),
    port: int = typer.Option(None, help="Port to run the server on"),
    browser: bool = typer.Option(True, help="Open the browser automatically."),
    webview: bool = typer.Option(True, help="Use the webview library if available."),
    verbose: bool = typer.Option(False, help="Run the server in verbose mode."),
):
    """ZnDraw: Visualize Molecules

    The ZnDraw CLI. Use 'zndraw version' to get the current version.
    """
    if not verbose:
        import logging

        log = logging.getLogger("werkzeug")
        log.setLevel(logging.ERROR)

    sys.path.insert(1, pathlib.Path.cwd().as_posix())
    if port is None:
        sock = socket.socket()
        sock.bind(("", 0))
        port = sock.getsockname()[1]
        sock.close()

    if file == "version":
        version_callback()

    if not pathlib.Path(file).exists():
        typer.echo(f"File {file} does not exist.")
        raise typer.Exit()

    globals.config = globals.Config(file=file)
    print(globals.config)

    if wv is not None and webview:
        wv.create_window("ZnDraw", app)
        with contextlib.suppress(wv.WebViewException):
            wv.start()
            return
    if browser:
        webbrowser.open(f"http://127.0.0.1:{port}")
    app.run(port=port)



import dataclasses
import socket
import threading
import webbrowser

from zndraw import app, globals


@dataclasses.dataclass
class ZnDraw:
    file: str
    port: int = None
    width: int = 700
    height: int = 600

    def __post_init__(self):
        if self.port is None:
            sock = socket.socket()
            sock.bind(("", 0))
            self.port = sock.getsockname()[1]
            sock.close()

        globals.config.file = self.file
        self.thread = threading.Thread(target=app.run, kwargs={"port": self.port})
        self.thread.start()
        webbrowser.open(f"http://127.0.0.1:{self.port}")

    def _repr_html_(self):
        from IPython.display import IFrame

        return IFrame(
            src=f"http://127.0.0.1:{self.port}", width=self.width, height=self.height
        )._repr_html_()

    def __del__(self):
        self.thread.join()
